﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using gudusoft.gsqlparser;
using gudusoft.gsqlparser.Units;
using System.IO;

namespace DependencySort
{
    enum ClauseType
    {
        where, connectby, startwith, orderby
    }

    class tableInExpr
    {

        private TLzCustomExpression expr;
        private dependencySort analysis;
        private TCustomSqlStatement stmt;
        public tableInExpr(dependencySort analysis, TLzCustomExpression expr, TCustomSqlStatement stmt)
        {
            this.stmt = stmt;
            this.analysis = analysis;
            this.expr = expr;
        }

        private bool expressionVisitor(TLz_Node pnode, Boolean pIsLeafNode)
        {
            TLzCustomExpression lcexpr = (TLzCustomExpression)pnode;

            if (lcexpr.oper == TLzOpType.Expr_FuncCall)
            {
                TLz_FuncCall func = (TLz_FuncCall)lcexpr.lexpr;
                if (func.args != null)
                {
                    for (int k = 0; k < func.args.Count(); k++)
                    {
                        if (func.args[k] is TLzCustomExpression)
                        {
                            TLzCustomExpression expr = (TLzCustomExpression)func.args[k];
                            expr.InOrderTraverse(expressionVisitor);
                        }
                    }
                }
                if (func._analyticFunction != null)
                {
                    TLz_List list = func._analyticFunction._partitionExprList;
                    tableInList(list, stmt);

                    if (func._analyticFunction._sortclause != null)
                    {
                        list = func._analyticFunction._sortclause.list1;
                        tableInList(list, stmt);

                        list = func._analyticFunction._sortclause.list2;
                        tableInList(list, stmt);
                    }
                }

            }
            else if (lcexpr.oper == TLzOpType.Expr_subquery)
            {
                if (lcexpr.lexpr is TSelectSqlStatement)
                {
                    analysis.searchSubQuery((TSelectSqlStatement)lcexpr.lexpr);
                }
            }
            else if (lcexpr.oper == TLzOpType.Expr_Case)
            {
                TLzCaseExpression expr = (TLzCaseExpression)lcexpr;
                TLzCustomExpression conditionExpr = expr.ConditionExpr;
                if (conditionExpr != null)
                {
                    conditionExpr.InOrderTraverse(expressionVisitor);
                }
                TLzCustomExpression defaultExpr = expr.DefaultExpr;
                if (defaultExpr != null)
                {
                    defaultExpr.InOrderTraverse(expressionVisitor);
                }
                TLz_List list = expr.WhenThenList;
                tableInList(list, stmt);
            }
            else if (lcexpr.oper == TLzOpType.Expr_Exists)
            {
                TLzCustomExpression expr = (TLzCustomExpression)lcexpr;
                if (expr.rexpr is TSelectSqlStatement)
                {
                    foreach (TLzField field in ((TSelectSqlStatement)expr.rexpr).Fields)
                    {
                        analysis.searchSubQuery((TSelectSqlStatement)expr.rexpr);
                    }
                }
            }
            return true;
        }

        private void tableInList(TLz_List list, TCustomSqlStatement stmt)
        {
            if (list != null)
            {
                for (int i = 0; i < list.Count(); i++)
                {
                    List<TLzCustomExpression> exprList = new List<TLzCustomExpression>();

                    if (list[i] is TLz_SortGroupBy)
                    {
                        exprList.Add((TLzCustomExpression)((TLz_SortGroupBy)list[i]).node);
                    }
                    else if (list[i] is TLzCustomExpression)
                    {
                        exprList.Add((TLzCustomExpression)list[i]);
                    }
                    else if (list[i] is TLzCaseWhen)
                    {
                        exprList.Add(((TLzCaseWhen)list[i]).WhenExpr);
                        exprList.Add(((TLzCaseWhen)list[i]).ThenExpr);
                    }

                    foreach (TLzCustomExpression lcexpr in exprList)
                    {
                        lcexpr.InOrderTraverse(expressionVisitor);
                    }
                }
            }
        }

        public void searchExpression()
        {
            this.expr.InOrderTraverse(expressionVisitor);
        }
    }



    public class dependencySort
    {
        private StringBuilder buffer = new StringBuilder();
        private Hashtable cteMap = new Hashtable();
        private List<String> tableList = new List<String>();
        private Hashtable viewMap = new Hashtable();
        private Hashtable viewTableMap = new Hashtable();
        private LinkedList<string> viewSortList = new LinkedList<string>();
        private List<TCustomSqlStatement> searchInSubQuerys = new List<TCustomSqlStatement>();
        private List<TCustomSqlStatement> searchInTables = new List<TCustomSqlStatement>();
        private List<TCustomSqlStatement> searchInClauses = new List<TCustomSqlStatement>();

        public string SortResult
        {
            get { return buffer.ToString(); }
        }

        static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                Console.WriteLine("Usage: {0} <sql script file path> <output file path>", Environment.GetCommandLineArgs()[0]);
                Console.WriteLine("sql script file path: The sql file will be sorted.");
                Console.WriteLine("output file path: Option, write the sorted result to the specified file.");
                //Console.Read();
                return;
            }

            string outputFile = null;
            StreamWriter writer = null;
            if (args.Length > 1)
            {
                outputFile = args[1];
            }

            if (outputFile != null)
            {
                try
                {
                    writer = new StreamWriter(outputFile);
                    Console.SetOut(writer);
                }
                finally { }
            }
            dependencySort sort = new dependencySort(new FileInfo(args[0]), TDbVendor.DbVMssql);
            Console.Write(sort.SortResult);
            //if (args.Length <= 1)
            //{
            //    Console.Read();
            //}
            //else
            {
                if (writer != null)
                {
                    writer.Close();
                }
            }
        }

        public dependencySort(string sql, TDbVendor dbVendor)
        {
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.SqlText.Text = sql;
            sortViews(sqlparser);
        }

        public dependencySort(FileInfo file, TDbVendor dbVendor)
        {
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.Sqlfilename = file.ToString();
            sortViews(sqlparser);
        }

        private void sortViews(TGSqlParser sqlparser)
        {
            int ret = sqlparser.Parse();

            if (ret != 0)
            {
                buffer.AppendLine(sqlparser.ErrorMessages);
                return;
            }

            analyzeSql(sqlparser);

            ArrayList sortList = new ArrayList();

            IEnumerator iter = viewSortList.GetEnumerator();
            while (iter.MoveNext())
            {
                String viewName = (String)iter.Current;
                sortView(sortList, viewName);
            }

            int index = 0;
            for (int i = 0; i < sqlparser.SqlStatements.Count(); i++)
            {
                if (sqlparser.SqlStatements[i] is TCreateViewSqlStatement)
                {
                    buffer.AppendLine(((TCreateViewSqlStatement)viewMap[sortList[index]]).AsText);
                    index++;
                }
                else
                {
                    buffer.AppendLine(sqlparser.SqlStatements[i].AsText);
                }
                buffer.AppendLine();
            }
        }

        private void sortView(ArrayList sortList, String viewName)
        {
            ArrayList tables = (ArrayList)viewTableMap[viewName];
            for (int i = 0; i < tables.Count; i++)
            {
                if (viewSortList.Contains((String)tables[i]))
                {
                    sortView(sortList, (String)tables[i]);
                }
            }
            if (!sortList.Contains(viewName))
                sortList.Add(viewName);
        }

        private void analyzeSql(TGSqlParser sqlparser)
        {
            for (int i = 0; i < sqlparser.SqlStatements.Count(); i++)
            {
                if (sqlparser.SqlStatements[i] is TCustomSqlStatement)
                {
                    TCustomSqlStatement stmt = (TCustomSqlStatement)sqlparser.SqlStatements[i];
                    if (stmt.cte != null && stmt.cte.ctes != null && stmt.cte.ctes.Count() > 0)
                    {
                        for (int j = 0; j < stmt.cte.ctes.Count(); j++)
                        {
                            TCommonTableExpression expression = (TCommonTableExpression)stmt.cte.ctes[j];
                            cteMap.Add(expression.ndTablename.AsText, expression.ndSelect_with_parens);
                        }
                    }
                    analyzeStatement(stmt);
                }
            }
        }

        private void analyzeStatement(TCustomSqlStatement stmt)
        {
            if (stmt is TCreateViewSqlStatement && ((TCreateViewSqlStatement)stmt).SubQuery != null)
            {
                TCreateViewSqlStatement viewStmt = (TCreateViewSqlStatement)stmt;
                tableList.Clear();
                analyzeStatement(viewStmt.SubQuery);
                string viewName = viewStmt.ViewName;
                viewName = escapeViewName(viewName);
                viewSortList.AddLast(viewName);
                viewTableMap.Add(viewName, new ArrayList(tableList));
                viewMap.Add(viewName, stmt);
            }
            else if (stmt is TSelectSqlStatement)
            {
                TSelectSqlStatement select = (TSelectSqlStatement)stmt;

                searchJoinFromStatement(select);

                if (select.SelectSetType != TSelectSetType.sltNone)
                {
                    analyzeStatement(select.LeftStmt);
                    analyzeStatement(select.RightStmt);
                }
                else
                {
                    foreach (TLzField field in select.Fields)
                    {
                        searchFields(field, select);
                    }
                }
            }
            else
            {
                foreach (TLzField field in stmt.Fields)
                {
                    searchFields(field, stmt);
                }
            }
        }

        private static string escapeViewName(string viewName)
        {
            string[] names = viewName.Split('.');
            StringBuilder buffer = new StringBuilder();
            for (int i = 0; i < names.Length; i++)
            {
                if (names[i][0] == '[' && names[i][names[i].Length - 1] == ']')
                    buffer.Append(names[i].Substring(1, names[i].Length - 2));
                else buffer.Append(names[i]);
                if (i < names.Length - 1)
                    buffer.Append(".");
            }
            viewName = buffer.ToString().ToLower();
            if (viewName[0] == '[' && viewName[viewName.Length - 1] == ']')
                viewName = viewName.Substring(1, viewName.Length - 2);
            return viewName;
        }

        private void searchJoinFromStatement(TSelectSqlStatement stmt)
        {
            if (stmt.JoinTables != null)
            {
                for (int i = 0; i < stmt.JoinTables.Count(); i++)
                {
                    TLzJoin join = stmt.JoinTables[i];
                    if (join.JoinItems != null)
                    {
                        for (int j = 0; j < join.JoinItems.Count(); j++)
                        {
                            TLzJoinItem joinItem = join.JoinItems[j];
                            TLzCustomExpression expr = joinItem.JoinQual;
                            searchExpression(expr, stmt);
                        }
                    }
                }
            }
        }

        public void searchFields(TLzField field, TCustomSqlStatement select)
        {
            TLzTableList tables = select.Tables;
            switch (field.FieldType)
            {
                case TLzFieldType.lftAttr:
                    searchTables(select);
                    searchClauses(select);
                    break;
                case TLzFieldType.lftExpression:

                    searchExpression(field.FieldExpr, select);
                    searchTables(select);
                    searchClauses(select);

                    if (field.FieldExpr.oper == TLzOpType.Expr_FuncCall)
                    {
                        TLz_FuncCall func = (TLz_FuncCall)field.FieldExpr.lexpr;
                        // buffer.AppendLine("function name {0}", func.funcname.AsText);

                        // check column in function arguments
                        if (func.args != null)
                        {
                            for (int k = 0; k < func.args.Count(); k++)
                            {
                                if (func.args[k] is TLzCustomExpression)
                                {
                                    TLzCustomExpression expr = (TLzCustomExpression)func.args[k];
                                    searchExpression(expr, select);
                                }
                            }
                        }
                        else
                        {
                            if (select.Tables[0].TableFullname != null)
                            {
                                if (!tableList.Contains(select.Tables[0].TableFullname.ToLower()))
                                {
                                    tableList.Add(select.Tables[0].TableFullname.ToLower());
                                }
                            }
                        }

                        if (func._analyticFunction != null)
                        {
                            TLz_List list = func._analyticFunction._partitionExprList;

                            searchExpressionList(select, list);

                            if (func._analyticFunction._sortclause != null)
                            {
                                list = func._analyticFunction._sortclause.list1;
                                searchExpressionList(select, list);

                                list = func._analyticFunction._sortclause.list2;
                                searchExpressionList(select, list);

                            }
                        }

                        // check order by clause
                        //if (select is TSelectSqlStatement && ((TSelectSqlStatement)select).GroupbyClause != null)
                        //{
                        //    for (int j = 0; j < ((TSelectSqlStatement)select).GroupbyClause.GroupItems.Count(); j++)
                        //    {
                        //        TLzGroupByItem i = (TLzGroupByItem)((TSelectSqlStatement)select).GroupbyClause.GroupItems[j];
                        //        searchExpression((TLzCustomExpression)i._ndExpr, select);
                        //        searchTables(select);
                        //    }

                        //}

                    }
                    break;
                case TLzFieldType.lftSubquery:
                    if (field.SubQuery is TSelectSqlStatement)
                    {
                        searchSubQuery((TSelectSqlStatement)field.SubQuery);
                    }
                    break;
                default:
                    buffer.AppendLine("searchFields of type: " + field.FieldType + " not implemented yet");
                    break;
            }
        }

        private void searchExpressionList(TCustomSqlStatement select, TLz_List list)
        {
            if (list == null)
                return;

            for (int i = 0; i < list.Count(); i++)
            {
                TLzCustomExpression lcexpr = null;
                if (list[i] is TLz_SortGroupBy)
                {
                    lcexpr = (TLzCustomExpression)((TLz_SortGroupBy)list[i]).node;
                }
                else if (list[i] is TLzCustomExpression)
                {
                    lcexpr = (TLzCustomExpression)list[i];
                }

                if (lcexpr != null)
                {
                    searchExpression(lcexpr, select);
                }
            }
        }

        public void searchSubQuery(TSelectSqlStatement select)
        {
            if (!searchInSubQuerys.Contains(select))
            {
                searchInSubQuerys.Add(select);
            }
            else
            {
                return;
            }

            searchJoinFromStatement(select);

            if (select.SelectSetType != TSelectSetType.sltNone)
            {
                searchSubQuery(select.LeftStmt);
                searchSubQuery(select.RightStmt);
            }
            else
            {
                foreach (TLzField field in select.Fields)
                {
                    searchFields(field, select);
                }
            }
        }

        private void searchClauses(TCustomSqlStatement select)
        {
            if (!searchInClauses.Contains(select))
            {
                searchInClauses.Add(select);
            }
            else
            {
                return;
            }
            if (select is TSelectSqlStatement)
            {

                TSelectSqlStatement statement = (TSelectSqlStatement)select;
                Hashtable clauseTable = new Hashtable();

                //if (statement.SortClause != null)
                //{
                //    TLzOrderByList sortList = (TLzOrderByList)statement.SortClause;
                //    for (int i = 0; i < sortList.Count(); i++)
                //    {
                //        TLzOrderBy orderBy = sortList[i];
                //        TLzCustomExpression expr = orderBy.SortExpr;
                //        clauseTable.Add(expr, ClauseType.orderby);
                //    }
                //}


                if (statement.WhereClause != null)
                {
                    clauseTable.Add((TLzCustomExpression)(statement.WhereClause), ClauseType.where);
                }
                //if (statement.ConnectByClause != null)
                //{
                //    clauseTable.Add((TLzCustomExpression)statement.ConnectByClause, ClauseType.connectby);
                //}
                //if (statement.StartwithClause != null)
                //{
                //    clauseTable.Add((TLzCustomExpression)statement.StartwithClause, ClauseType.startwith);
                //}
                foreach (TLzCustomExpression expr in clauseTable.Keys)
                {
                    ClauseType type = (ClauseType)clauseTable[expr];
                    searchExpression(expr, select);
                    searchTables(select);

                }
            }
        }

        void searchTables(TCustomSqlStatement select)
        {
            if (!searchInTables.Contains(select))
            {
                searchInTables.Add(select);
            }
            else
            {
                return;
            }

            TLzTableList tables = select.Tables;

            if (tables.Count() == 1)
            {
                TLzTable lzTable = tables[0];
                if ((lzTable.TableType == TLzTableType.lttAttr)
                    && (lzTable.TableAlias.Length == 0))
                {
                    if (cteMap.ContainsKey(lzTable.Name))
                    {
                        searchSubQuery((TSelectSqlStatement)cteMap[lzTable.Name]);
                    }
                    else
                    {
                        if (lzTable.TableFullname != null)
                        {
                            if (!tableList.Contains(lzTable.TableFullname.ToLower()))
                            {
                                tableList.Add(lzTable.TableFullname.ToLower());
                            }
                        }
                    }
                }
            }

            foreach (TLzTable lztable in tables)
            {
                switch (lztable.TableType)
                {
                    case TLzTableType.lttAttr:
                        if (cteMap.ContainsKey(lztable.Name))
                        {
                            searchSubQuery((TSelectSqlStatement)cteMap[lztable.Name]);
                        }
                        else
                        {
                            if (lztable.TableFullname != null)
                            {
                                if (!tableList.Contains(lztable.TableFullname.ToLower()))
                                {
                                    tableList.Add(lztable.TableFullname.ToLower());
                                }
                            }
                        }
                        break;
                    case TLzTableType.lttSubquery:
                        searchSubQuery((TSelectSqlStatement)lztable.SubQuery);
                        break;
                    default:
                        break;
                }
            }
        }

        void searchExpression(TLzCustomExpression expr, TCustomSqlStatement stmt)
        {
            tableInExpr c = new tableInExpr(this, expr, stmt);
            c.searchExpression();
        }
    }
}
